-- This addon was started from Aanye XP as a base, which uses the MIT License. Thanks to Aanye for the code I started with.
-- This addon also incorporates code from Broker Cash, which also uses the MIT License. Thansk Septh!

-- Local variables --

local HasActiveAzeriteItem, FindActiveAzeriteItem, GetAzeriteItemXPInfo, GetPowerLevel = C_AzeriteItem.HasActiveAzeriteItem, C_AzeriteItem.FindActiveAzeriteItem, C_AzeriteItem.GetAzeriteItemXPInfo, C_AzeriteItem.GetPowerLevel
local addonName, addonSpace = ...
local L = LibStub('AceLocale-3.0'):GetLocale(addonName)

local VERSION = GetAddOnMetadata(addonName, 'Version')

local libLDB  = LibStub('LibDataBroker-1.1')
local libQTip = LibStub('LibQTip-1.0')
local libRealmInfo = LibStub("LibRealmInfo")

local db
local currentXp, currentMaxXp, startXp
local currentLevel, startLevel
local lastXpGain=0
local azeriteItemLocation
local azeriteIcon = "Interface/Icons/Inv_smallazeriteshard"
local azeriteItemIcon = azeriteIcon
local ap = {}
local COLOR_GREEN  = CreateColor(0.1, 0.8, 0.1, 1)
local apmaxes =    {350,400, 450, 500, 550, 600, 650, 700,1050,1580,2370, 3560, 5350, 8000,10400,13500,17550,22800, 29650, 38600, 50100, 65150, 84700,110100,143150,186100, 241950, 314550, 408900, 531550, 691000, 898300}
local aptotals = {0,350,750,1200,1700,2250,2850,3500,4200,5250,6830,9200,12760,18110,26110,36510,50010,67560,90360,120010,158610,208710,273860,358560,468660,611810,797910,1039860,1354410,1763310,2294860,2985860,3884160}

local tth = select(2, GameTooltipText:GetFontObject():GetFont())
local AZERITE_ICON_STRING = ('|T'..azeriteIcon..':%d:%d:2:0|t'):format(tth, tth)

local f = CreateFrame("frame")

local function GetRegionStartOfWeek()
  local region = libRealmInfo:GetCurrentRegion()
  if region == "US" then
    return 3
  elseif region == "EU" then
    return 4
  else
    return 3 -- FIXME
  end
end

local sv_defaults = {
	global = {
		ldb = {
			display_format = "curAll",
			display_level = true,
			display_gain = true,
      display_chars = true,
      start_of_week = GetRegionStartOfWeek()
		}
	},
  char = {
    lastSaved  = 0,
    powerLevel = 0,
    xp         = 0,
    maxXp      = 0,

    session    = 0,
    day        = 0,
    week       = 0,
    lastweek   = 0,
    ever       = 0,

    sources    = {
      questIDs = {},  
      uiMapIDs = {}
    },

    days = {}
  }
}
local intervals = { 'session', 'day', 'week', 'ever' } -- not including last week
local options_panel = {
  name    = ('%s v%s'):format(addonName, VERSION),
  handler = f,
  type    = 'group',
  childGroups = 'tab',
  args = {
    options = {
      name  = L["Options"],
      type  = 'group',
      order = 1,
      get   = 'OptionsPanel_GetOpt',
      set   = 'OptionsPanel_SetOpt',
      args  = {
        ldb = {
          name   = L["LDB Display"],
          type   = 'group',
          inline = true,
          order  = 10,
          args   = {
            display_format = {
              name  = L["Text Format"],
              desc  = L["How to format the text shown in the broker display."],
              style = 'dropdown',
              values = {
              	curAll = L["Current/Maximum (Current/Maximum%)"],
              	remAll = L["Remaining (Remaining%)"],
              	curPer = L["Current/Maximum%"],
              	remPer = L["Remaining%"]
              },
              width = 'double',
              type  = 'select',
              order = 1,
            },
            display_level = {
              name  = L["Display Level"],
              desc  = L["Display the Azerite item level in the broker display."],
              descStyle = 'inline',
              type  = 'toggle',
              width = 'full',
              order = 2,
            },
            display_gain = {
              name  = L["Display AP Gain"],
              desc  = L["Display the last AP gain in the broker display."],
              descStyle = 'inline',
              type  = 'toggle',
              width = 'full',
              order = 3,
            },
            display_chars = {
              name  = L["Display Other Characters"],
              desc  = L["Display Azerite Item data for other characters in the tooltip."],
              descStyle = 'inline',
              type  = 'toggle',
              width = 'full',
              order = 4,
            },
            start_of_week =
            {
              name = L["Week Day"],
              desc = L["Select the day of the week."],
              order = 5,
              type = "select",
              values = {WEEKDAY_SUNDAY, WEEKDAY_MONDAY, WEEKDAY_TUESDAY, WEEKDAY_WEDNESDAY, WEEKDAY_THURSDAY, WEEKDAY_FRIDAY, WEEKDAY_SATURDAY}
            },
          },
        },
      },
    },
  }
}

LibStub('AceConfig-3.0'):RegisterOptionsTable(addonName, options_panel)
local optionsFrame = LibStub('AceConfigDialog-3.0'):AddToBlizOptions(addonName)

-- Instantiate Frame/Object --
f:SetScript("OnEvent", function(self, event, ...) if self[event] then return self[event](self, event, ...) end end)
local dataobj = LibStub("LibDataBroker-1.1"):NewDataObject("Gallywix_Azerite_Ledger", {type = "data source", text = " ", icon = azeriteItemIcon})

-- Frame functions --

function f:PLAYER_ENTERING_WORLD(event, ...)
	local isInitialLogin, isReloadingUi = ...

	db = LibStub("AceDB-3.0"):New("Gallywix_Azerite_Ledger_DB", sv_defaults, true)

	if isInitialLogin then
		db.char.session = 0
	end

	self:RegisterEvent("AZERITE_ITEM_EXPERIENCE_CHANGED")
	self:RegisterEvent("AZERITE_ITEM_POWER_LEVEL_CHANGED")

	self:UnregisterEvent("PLAYER_ENTERING_WORLD")
	self.PLAYER_ENTERING_WORLD = nil
	
	self:SetAzeriteVars()
	self:SetBrokerText()
  self:RecordXpGain(0)
end

function f:SetAzeriteVars()
	if not HasActiveAzeriteItem() then
		return
	end
	
	if not azeriteItemLocation then 
		-- set up start values
		azeriteItemLocation = FindActiveAzeriteItem()
		azeriteItemIcon = C_Item.GetItemIcon(azeriteItemLocation)
		startXp = GetAzeriteItemXPInfo(azeriteItemLocation)
		startLevel, currentMaxXp = GetPowerLevel(azeriteItemLocation)

    if db.char.ever == 0 then
      -- estimate total AP earned previously based on current level/xp
      db.char.ever = aptotals[startLevel] + startXp
    end
	end

	currentXp, currentMaxXp = GetAzeriteItemXPInfo(azeriteItemLocation)
	currentLevel = GetPowerLevel(azeriteItemLocation)

	ap[currentLevel] = ap[currentLevel] or currentMaxXp
end

function f:SetBrokerText()
	local dotext
	dataobj.icon = azeriteItemIcon
  if not HasActiveAzeriteItem() then
		dotext = L["No Active Azerite Item"]
	else
		-- Show azerite item current level
		if db.global.ldb.display_level == true and currentLevel ~= nil then
			dotext = currentLevel..": "
		else
			dotext = ""
		end

		if db.global.ldb.display_format == "curAll" then
			dotext = dotext..string.format("%d/%d (%d%%)", currentXp, currentMaxXp, currentXp/currentMaxXp*100)
		elseif db.global.ldb.display_format == "remAll" then
			dotext = dotext..string.format("%d (%d%%)", (currentMaxXp-currentXp), (currentMaxXp-currentXp)/currentMaxXp*100)
		elseif db.global.ldb.display_format == "curPer" then
			dotext = dotext..string.format("%01.1f%%", currentXp/currentMaxXp*100)
		elseif db.global.ldb.display_format == "remPer" then
			dotext = dotext..string.format("%01.1f%%", (currentMaxXp-currentXp)/currentMaxXp*100)
		end

		if db.global.ldb.display_gain == true and lastXpGain >= 0 then
			dotext = dotext..COLOR_GREEN:WrapTextInColorCode(string.format(" +%d", lastXpGain))
		end
	end
	dataobj.text = ' '..dotext..' '
end

function f:RecordXpGain(gain)
  for _,stat in ipairs( intervals ) do
      db.char[stat] = db.char[stat] + gain
  end
  db.char.powerLevel = currentLevel
  db.char.xp = currentXp
  db.char.maxXp = currentMaxXp
  db.char.lastSaved = time()

  local current_map = C_Map.GetBestMapForUnit("player")
  db.char.sources.uiMapIDs[current_map] = db.char.sources.uiMapIDs[current_map] or 0
  db.char.sources.uiMapIDs[current_map] = db.char.sources.uiMapIDs[current_map] + gain

  local now = date('*t')
  local today = string.format("%d.%02d.%02d", now.year, now.month, now.day)
  db.char.days[today] = db.char.days[today] or {}
  db.char.days[today].yday = now.yday
  db.char.days[today].ap = db.char.days[today].ap or 0
  db.char.days[today].ap = db.char.days[today].ap + gain
end

-- from Broker: Cash
-------------------------------------------------------------------------------
function f:CheckStatsResets()
    local now = date('*t')

    ---------------------------------------------------------------------------
    -- 1/ Compute reset dates for stats
    ---------------------------------------------------------------------------
    local startOfDay, startOfMonth, startOfYear, startOfWeek
    local reset = {
        hour = 0,   -- Toutes les limites sont fixées à 00:00:00
        min  = 0,   -- TODO: Gérer l'heure d'été/hiver ? Si oui, comment ?
        sec  = 0
    }

    -- Start of today
    reset.day   = now.day
    reset.month = now.month
    reset.year  = now.year
    startOfDay = time(reset)

    reset.day    = 1
    startOfMonth = time(reset)

    reset.month = 1
    startOfYear = time(reset)

    -- Start of current week
    local numDaysBack = now.wday - FIRST_DAY_OF_WEEK
    if numDaysBack < 0 then
        numDaysBack = 7 - now.wday
    end
    reset.day   = now.day - numDaysBack
    reset.month = now.month
    reset.year  = now.year
    if reset.day < 1 then
        reset.month = reset.month - 1
        if reset.month < 1 then
            reset.month = 12
            reset.year = reset.year - 1
        end
        -- # of days in month (cf. http://lua-users.org/wiki/DayOfWeekAndDaysInMonthExample)
        reset.day = date('*t', time( { day = 0, month = reset.month + 1, year = reset.year } ))['day']
    end
    startOfWeek = time(reset)

    ---------------------------------------------------------------------------
    -- 2/ Reinit old stats?
    ---------------------------------------------------------------------------
    for key,char in pairs(self.sv.char) do
        local lastSaved, resetValue = char.lastSaved or 0, key == currentCharKey and 0 or nil

        if lastSaved < startOfDay   then char.day   = resetValue end
        if lastSaved < startOfWeek  then char.week  = resetValue end
        if lastSaved < startOfMonth then char.month = resetValue end
        if lastSaved < startOfYear  then char.year  = resetValue end
    end

    ---------------------------------------------------------------------------
    -- 3/ Recompute realms?
    ---------------------------------------------------------------------------
    self:AuditRealms()

    -- Rafraîchit le tooltip s'il est affiché
    --if mainTooltip and mainTooltip:IsShown() then
    --    self:UpdateMainTooltip()
    --end

    ---------------------------------------------------------------------------
    -- 4/ Relance le chronomètre jusqu'à demain minuit pour la prochaine vérification
    ---------------------------------------------------------------------------
    now.day  = now.day + 1	-- Demain
    now.hour = 0			-- à 0 heure
    now.min  = 0			-- 0 minute
    now.sec  = 1			-- et 1 seconde (marge de sécurité)
    self:ScheduleTimer('CheckStatsResets', difftime(time(now), time()))
end


function f:ShowOptionsPanel(msg)
    if InCombatLockdown() or GameMenuFrame:IsShown() or InterfaceOptionsFrame:IsShown() then return end

    LibStub('AceConfigDialog-3.0'):Open(addonName)
end

function f:ToggleOptionsPanel()
    local acd = LibStub('AceConfigDialog-3.0')
    if acd.OpenFrames[addonName] then
      acd:Close(addonName)
    else
      f:ShowOptionsPanel()
    end
end

function f:OptionsPanel_GetOpt(info)
    return db.global[info[#info-1]][info[#info]]
end
function f:OptionsPanel_SetOpt(info, value)
    -- info[#info - 1] = 'ldb', info[#info] = option checked
    db.global[info[#info-1]][info[#info]] = value
    self:SetBrokerText()
end

-- Events --

function f:AZERITE_ITEM_EXPERIENCE_CHANGED(event, ...)
	local azeriteItemLocation, oldExperienceAmount, newExperienceAmount = ...
	--DEFAULT_CHAT_FRAME:AddMessage(string.format("GAL AZERITE_ITEM_EXPERIENCE_CHANGED old AP: %d new AP: %d - Level %d", oldExperienceAmount, newExperienceAmount, GetPowerLevel(azeriteItemLocation)))
	lastXpGain = newExperienceAmount - oldExperienceAmount

  -- if it's not equal then we will defer the xp gain to the power level event
	if GetPowerLevel(azeriteItemLocation) == currentLevel then 
    self:SetAzeriteVars()
    self:SetBrokerText()
    self:RecordXpGain(lastXpGain)
	end
end

-- it is assumed that this event always happens AFTER a XP change event
function f:AZERITE_ITEM_POWER_LEVEL_CHANGED(event, ...)
	local azeriteItemLocation, oldPowerLevel, newPowerLevel, unlockedEmpoweredItemsInfo = ...
	local lastLevelGain = newPowerLevel - oldPowerLevel
	if lastLevelGain ~= 1 then
		DEFAULT_CHAT_FRAME:AddMessage("Gallywix's Azerite Ledger: Unexpected Azerite Item level change of %d, please report as a bug", lastLevelGain)
	end

	-- Correct our power gain 
	lastXpGain = lastXpGain + ap[oldPowerLevel]

  --DEFAULT_CHAT_FRAME:AddMessage(string.format("GAL AZERITE_ITEM_POWER_LEVEL_CHANGED old: %d new: %d - AP gain: %d", oldPowerLevel, newPowerLevel, lastXpGain))
  self:SetAzeriteVars()
  self:SetBrokerText()

  self:RecordXpGain(lastXpGain)
end

-- Local Functions --

local function GetTipAnchor(frame)
	local x,y = frame:GetCenter()
	if not x or not y then return "TOPLEFT", "BOTTOMLEFT" end
	local hhalf = (x > UIParent:GetWidth()*2/3) and "RIGHT" or (x < UIParent:GetWidth()/3) and "LEFT" or ""
	local vhalf = (y > UIParent:GetHeight()/2) and "TOP" or "BOTTOM"
	return vhalf..hhalf, frame, (vhalf == "TOP" and "BOTTOM" or "TOP")..hhalf
end

-- OnClick --

function dataobj.OnClick(self,button)
	if IsShiftKeyDown() then
		f.ToggleOptionsPanel()
		return
	end
  if button == "LeftButton" then
  	db.global.ldb.display_level = not db.global.ldb.display_level
	elseif button == "RightButton" then
		db.global.ldb.display_gain = not db.global.ldb.display_gain
  elseif button == "MiddleButton" then
    --TestApGain()
	end
	f.SetBrokerText()
	dataobj.SetTooltipContents()
end

-- Tooltip --

function dataobj.SetTooltipContents()
	GameTooltip:ClearLines()

	GameTooltip:AddLine(AZERITE_ICON_STRING.." "..L["Azerite Earned Summary"])
	GameTooltip:AddLine(" ")
  if not HasActiveAzeriteItem() then
		GameTooltip:AddLine(L["No Active Azerite Item"],1,1,1)
	else
		GameTooltip:AddDoubleLine(L["Current Azerite"]..":", currentXp.."/"..currentMaxXp, nil,nil,nil, 1,1,1)
		GameTooltip:AddDoubleLine(L["Remaining to Level"]..":", currentMaxXp-currentXp, nil,nil,nil, 1,1,1)

		GameTooltip:AddDoubleLine(L["Session"]..":",     db.char.session,  nil,nil,nil, 0.2,1,0.2)
    GameTooltip:AddDoubleLine(L["Today"]..":",       db.char.day,      nil,nil,nil, 1,1,1)
    GameTooltip:AddDoubleLine(L["This week"]..":",   db.char.week,     nil,nil,nil, 1,1,1)
    GameTooltip:AddDoubleLine(L["Last week"]..":",   db.char.lastweek, nil,nil,nil, 1,1,1)
    GameTooltip:AddDoubleLine(L["All time"]..":",    db.char.ever,     nil,nil,nil, 1,1,1)
	end

  if db.global.ldb.display_chars then
    GameTooltip:AddLine(" ")
    for key,char in pairs(db.sv.char) do
      if char.powerLevel and char.powerLevel >= 1 then
        GameTooltip:AddDoubleLine(key, string.format("%d: %d/%d", char.powerLevel, char.xp, char.maxXp), nil,nil,nil, 1,1,1)
      end
    end
  end

  GameTooltip:AddLine(" ")
  GameTooltip:AddLine(L["|cffeda55fShift-Click|r to show options."], 0.2, 1, 0.2, 1)

	GameTooltip:Show()
end

local function SplitCharKey(charKey)
    local charName, charRealm = ('-'):split(charKey, 2)
    return charName:trim(), charRealm:trim()
end

function dataobj.OnLeave() GameTooltip:Hide() end
function dataobj.OnEnter(self)
 	GameTooltip:SetOwner(self, "ANCHOR_NONE")
	GameTooltip:SetPoint(GetTipAnchor(self))
	dataobj.SetTooltipContents()
end

-- Startup Events --
LibStub('AceConfig-3.0'):RegisterOptionsTable(addonName, options_panel)
if IsLoggedIn() then f:PLAYER_ENTERING_WORLD() else f:RegisterEvent("PLAYER_ENTERING_WORLD") end
